package ruon.jiao;
/**
 * This API is the property of R-U-ON LLC
 * You can embed this API in other applications and freely distribute it.
 * R-U-ON maintains the right to block agents that create an extreme load
 * on the server.
 */

import java.io.IOException;
import java.util.*;
import java.text.MessageFormat;


/**
 * The <code>JIAO</code> class is an API to the IAOP protocol, allowing you
 * to communicate with the R-U-ON platform.
 * <p>
 * A process using this class is called an Agent. There are two kind of agents,
 * Private and Public. A public agent is downloaded from the R-U-ON site and
 * is tightly integrated with the R-U-ON web application.
 * A private Agent is developed by an R-U-ON user. A private Agent can
 * be monitored by the R-U-ON system but has a few limitations. It cannot be downloaded
 * from the R-U-ON site and therefor cannot be automatically upgraded. In addition it
 * cannot be configured by the R-U-ON system (embedded configuration variables are
 * added to the agent during the download process).
 * Look at our agent publishing program to see how you can make your agent an integral part
 * of the R-U-ON system.
 * <p>
 * When creating an instance the caller should pass the Agent type and version. The JIAO object
 * should be created once for the lifetime of the agent. It must not be constructed each time
 * a call is made as that will seem to the R-U-ON platform like an agent reboot.
 * <p>
 * When the calling process is running for the first time it should obtain
 * a new instance id by invoking <code>register()</code> and passing the customer id.
 * The agent id should be persisted for subsequent execution of the agent process.
 * <p>
 * After the first construction the caller should use <code>setAgentId()</code> to set the id to
 * the persisted value.
 * If the caller uses register every time, a new agent will be created each time the agent is executed.
 * <p>
 * In order to make sure the server considers the Agent alive, the Agent
 * must call <code>keepalive()</code> and act according to the returned Directive. The most
 * common Directive is a request to SLEEP. The duration of the SLEEP indicates
 * when the server is expecting the  agent to make contact again (by callin
 * <code>keepalive()</code>  again).
 * <p>
 * To report alarms use the <code>reportAlarms()</code> method.
 * <p>
 * For information about the IAOP protcol, goto <a href="http://www.r-u-on.com/api/IAOP.pdf">http://www.r-u-on.com/api/IAOP.pdf</a>
 *
 * @author  R-U-ON
 * @version 1.0, 6/29/06
 * @since   jiao 1.0
 */


public class JIAO {

   /**
    * Creates the JIAO class and initialized the agent type and version.
    *
    * @param   agentType    The agent's type. This name will be displayed as the agent type in the management screen.
    * @param   agentVersion The agent's version. In the case of a public agent, the version string is used to determine if the agent should upgrade.
    */
    public JIAO(String agentType, String agentVersion) {
        this.agentVersion = agentVersion;
        this.agentType = agentType;
        iaop = new IAOP();
    }

    /**
     * Registers a new agent with the R-U-ON service.<p>
     * This method should be called once per agent to obtain a new agent id. The agent id should be
     * persisted for subsequent uses. Do not call this function each time the process is launched
     * as each call will create a new agent for the given customer.
     * <p>
     * This method is intended for the use of a public agent.
     *
     * @return The new agent id, to be persisted for future JIAO instances.
     * @throws IOException
     *         If an IO error has occored.
     * @throws JIAOException
     *          If an IAOP error has occored
     */
    public String register() throws IOException, JIAOException {
        return register(getEmbeddedVariable("ACCOUNT_ID"));
    }

    /**
     * Registers a new agent with the R-U-ON service.<p>
     * This method should be called once per agent to obtain a new agent id. The agent id should be
     * persisted for subsequent uses. Do not call this function each time the process is launched
     * as each call will create a new agent for the given customer.
     * <p>
     * This method is intended for the use of a private agent.
     *
     * @param accountId The R-U-ON account id (obtained from the account settings page in the R-U-ON site).
     * @return The new agent id, to be persisted for future JIAO instances.
     * @throws IOException
     *         If an IO error has occored.
     * @throws JIAOException
     *          If an IAOP error has occored
     */
    public String register(String accountId) throws IOException, JIAOException {
        if (accountId==null) {
            throw new JIAOException("Account id cannot be null");
        }
        id = accountId;
        Response r = iaop.request(this, "<register />");

        Directive d = internalAction(r, false);
        if (d!=null) {
            if (d.getVerb()==Directive.DIE) {
                throw new JIAOException("Error - " + d.getX());
            }
        }

        id = r.id();
        if (id==null) {
            throw new JIAOException("Unexpeceted result: "+r);
        }
        return id;
    }

    /**
     * Sets the agent id with the value persisted from the call to register()
     * <p>
     * When register succeeds, it is not neccessary to call this method.
     * @param agentId The agent id returned by register in the past.
     */
    public void setAgentId(String agentId) {
        this.id = agentId;
    }

    /**
     * Directive is returned by methods communicating with the server to indicate
     * what the server is expecting the agent to do.
     * A directive has a verb which describes the action and a modifier called X
     * that changes in meaning, depending on the verb.
     */
    public static class Directive {
        static class VERB {
            String s;
            static List l = new ArrayList();
            VERB(String s) {
                this.s = s;
                l.add(this);
            }

            /**
             * Returns a String representation
             * @return The String
             */
            public String toString() {
                return s;
            }
        }

        /**
         * Agent is expected call keep alive again in X seconds.
         * The agent can communicate alarms in the mean time. If
         * a new sleep directive is obtained the original one is
         * discarded (i.e. the sleep counter resets).
         */
        public static VERB SLEEP=new VERB("sleep");
        /**
         * Agent should refresh its configuration by calling
         * configure.
         * This directive is a result of the user updating
         * the configuration.
         */
        public static VERB CONFIGURE=new VERB("config");
        /**
         * This directive is intended for public agents only
         * and tells the agent to perform an upgrade.
         * Private agents cannot be upgraded via the R-U-ON
         * platform.
         * X is the text associated with this request.
         */
        public static VERB UPGRADE=new VERB("upgrade");
        /**
         * Sent as a response to an explicit request by
         * the user to perform an uninstall.
         * X is the text associated with this request.
         */
        public static VERB UNINSTALL=new VERB("uninstall");
        /**
         * Something is going terribly wrong and the server
         * would like the agent to stop communicating.
         * After sending back this directive the JIAO object
         * will no longer be valid and will throw exceptions
         * for any subsequent request to communicate with the
         * server.
         * X is the reason associated with this request.
         */
        public static VERB DIE=new VERB("die");

        /**
         * Return the directive's verb
         * @return The verb Object
         */
        public VERB getVerb() {
            return verb;
        }

        /**
         * Verb modifier, meaning depends on verb context
         * @return The modifier
         */
        public String getX() {
            return x;
        }
        /**
         * Returns a String representation
         * @return The String
         */
        public String toString() {
            return new StringBuffer().append(getVerb()).append(":").append(getX()).toString();
        }


        ///////////////// PRIVATE ////////////////////
        private VERB verb;
        private String x;

        private Directive(String directive) {
            String [] a = directive.split(":",2);

            this.verb = verb(a[0]);
            this.x = a[1];
        }
        private final static Map vm = new HashMap();
        private VERB verb(String s) {
            synchronized(vm) {
                if (vm.size()==0) {
                    Iterator i = VERB.l.iterator();
                    while (i.hasNext()) {
                        VERB v =  (VERB) i.next();
                        vm.put(v.s, v);
                    }
                }
            }
            return ((VERB) vm.get(s));
        }
    }

    /**
     * Send a keep-alive signal to the R-U-ON server.
     * @return An instruction that should be executed by the agent.
     * @throws IOException
     *         If an IO error has occored.
     * @throws JIAOException
     *          If an IAOP error has occored
     * @see     JIAO.Directive
     */
    public Directive keepAlive() throws IOException, JIAOException {
        Response r = iaop.request(this, "<keepalive />");
        internalAction(r, true);
        return new Directive(r.directive());
    }

    /**
     * A class representing an Alarm, a Clear instructions or an Event.
     * Passed to the reportAlarms method.
     * @see JIAO
     */
    public static abstract class  IAlarm {
        /**
         * Represents the severity of an Alarm or Event
         */
        public static class Severity{
            String s;
            private Severity(String s) {
                this.s = s;
            }
            /**
             * Returns a String representation
             * @return The String
             */
            public String toString() {
                return s;
            }
        }

        /**
         * Critical Alarm or Event
         */
        public static final Severity CRITICAL = new Severity("C");
        /**
         * Major Alarm or Event
         */
        public static final Severity MAJOR = new Severity("M");
        /**
         * Minor Alarm or Event
         */
        public static final Severity MINOR = new Severity("m");

        private final static String specialChars[][] = {
            { "&amp;",  "&" },
            { "&quot;", "\""},
            { "&lt;",    "<" },
            { "&gt;",    ">" },
        };
        protected String xmlSafe(String resource) {
            for (int i = 0; i < specialChars.length; i++) {
                String[] pair = specialChars[i];
                resource = resource.replaceAll(pair[1], pair[0]);
            }
            return resource;
        }


        abstract String tag();
    }

    /**
     * An Alarm object passed to the reportAlarms() method to indicate an alarm
     */
    public static class Alarm extends IAlarm {
        private Object [] args;

        /**
         * Create an Alarm object
         * @param resource  Resource that has the alarm. Combined with id, this is the unique identifier of the alarm. (Up to 256 bytes)
         * @param id        An id that is unique for the alarmed resource. (Up to 256 bytes)
         * @param severity  The severity of the alarm (Critical, Major or Minor)
         * @param description The verbal description of the alarm (Up to 1024 bytes)
         */
        public Alarm(String resource, String id, IAlarm.Severity severity, String description) {
            args = new Object[]{xmlSafe(resource), id, severity,description};
        }

        String tag() {
            return MessageFormat.format("<alarm resource=\"{0}\" id=\"{1}\" severity=\"{2}\"><![CDATA[{3}]]></alarm>", args);
        }
    }

    /**
     * A Clear Alarm instruction passed to the reportAlarms() method.<p>
     * The combination of resource and id uniquly identifies the alarm to be cleared.
     */
    public static class Clear extends IAlarm {
        private Object [] args;

        /**
         * Create a Clear object
         * @param resource The alarmed resource
         * @param id The id of the alarm
         */
        public Clear(String resource, String id) {
            args = new Object[]{resource, id};
        }
        /**
         * Create a Clear object
         * @param resource The alarmed resource
         * @param id The id of the alarm
         * @param description The description to be used in notification for the clear event. If this
         * is not specified, the system will use a generic statement to inform the user of the alarm
         * being cleared.
         */
        public Clear(String resource, String id, String description) {
            args = new Object[]{xmlSafe(resource), id, description};
        }
        String tag() {
            if (args.length==3) {
                return MessageFormat.format("<clear resource=\"{0}\" id=\"{1}\" ><![CDATA[{2}]]></clear>", args);
            } else {
                return MessageFormat.format("<clear resource=\"{0}\" id=\"{1}\" />", args);
            }
        }
    }

    /**
     * An Event object, passed to the reportAlarms() method.<p>
     * An Event is reported to the user just like an alarm but is not maintained in the
     * active alarm table and is not expected to have a corresponding Clear event.
     */
    public static class Event extends IAlarm {
        private Object [] args;
        /**
         * Create an Event object
         * @param resource  Resource that has the alarm (Combined with id, this is the unique identifier of the alarm). (Up to 256 bytes)
         * @param id        An id that is unique for the alarmed resource. (Up to 256 bytes)
         * @param severity  The severity of the alarm (Criticalm, Major or Minor).
         * @param description The verbal description of the alarm (Up to 1024 bytes).
         */
        public Event(String resource, String id, IAlarm.Severity severity, String description) {
            args = new Object[]{xmlSafe(resource), id, severity,description};
        }


        String tag() {
            return MessageFormat.format("<event resource=\"{0}\" id=\"{1}\" severity=\"{2}\"><![CDATA[{3}]]></event>", args);
        }
    }

    /**
     * Reports alarms to the R-U-ON server.<p>
     *
     * @param alarms An array of alarms, clears and events (Alarm, Clear, Event)
     * @param incremental Indicates if the call is the full list of active alarms (snapshot) or a list of
     * alarms that should be added or delted. In the case of a snapshot, all active alarms that are not reported
     * in the array are implicitly supressed.
     * @return An instruction that should be executed by the agent.
     * @throws IOException
     *         If an IO error has occored.
     * @throws JIAOException
     *          If an IAOP error has occored
     * @see     JIAO.Alarm
     * @see     JIAO.Clear
     * @see     JIAO.Event
     */
    public Directive reportAlarms(IAlarm [] alarms, boolean incremental) throws IOException, JIAOException {
        StringBuffer sb = new StringBuffer();
        sb.append("<alarms incremental=\"");
        sb.append(incremental);
        sb.append("\">");
        for (int i=0;i<alarms.length;++i) {
            sb.append(alarms[i].tag());
        }
        sb.append("</alarms>");

        Response r = iaop.request(this, sb.toString());
        internalAction(r, true);
        return new Directive(r.directive());
    }

    /**
     * Set configuration meta data and retrieve current configuration.<p>
     *
     * @param metaConfig A MetaConfig object, representing the meta data.
     * @return A Config object representing the current agent configuration
     * @throws IOException
     *         If an IO error has occored.
     * @throws JIAOException
     *          If an IAOP error has occored
     * @see     MetaConfig, Config
     */
    public Config configure(MetaConfig metaConfig) throws IOException, JIAOException {

        Response r = iaop.request(this, metaConfig.toXML());

        return new Config(r.data);
    }

    public Directive setParams(AgentParams params) throws IOException, JIAOException {

        Response r = iaop.request(this, params.toString());
        internalAction(r, true);
        return new Directive(r.directive());
    }

    /**
     * Only used in public agents.<p>
     * Should be used as a response to an upgrade directive.<p>
     * Contacts the server and get's an image of the agent binary.
     *
     * @return The binary image of the agent.
     * @throws IOException
     *         If an IO error has occored.
     * @throws JIAOException
     *          If an IAOP error has occored
     */
    public byte [] getBinary() throws JIAOException, IOException {
        throw new JIAOException("Not supported for private agents (yet)");
    }

    /**
     * Only used in public agents.<p>
     * Read variables embedded in the agent during the download process.
     * @param name Name of the variable
     * @return  The value of the variable
     * @throws JIAOException    Thrown when the variable is not found
     */
    public String getEmbeddedVariable(String name) throws JIAOException {
        throw new JIAOException("Not supported for private agents (yet)");
    }

    ///////////////////////// PRIVATE ///////////////////////
    String agentType;
    String agentVersion;
    String id;
    IAOP iaop;

    private Directive internalAction(Response r, boolean strict) throws JIAOException {
        String err = r.error();
        if (err!=null) {
            throw new JIAOException(err);
        }
        String ddd = r.directive();
        if (ddd==null) {
            return null;
        }
        Directive d = new Directive(ddd);
        if (d.getVerb()==Directive.DIE) {
            if (strict) {
                iaop = null;
            }
        }
        return d;
    }

}
